import torch

from mipcandy.common import Normalize


def convert_ids_to_logits(ids: torch.Tensor, num_classes: int, *, channel_dim: int = 1) -> torch.Tensor:
    """
    :param ids: class ids (..., 1, ...)
    :param num_classes: number of classes
    :param channel_dim: the index of the channel dimension
    :return: logits (..., num_classes, ...)
    """
    if torch.is_floating_point(ids) or (ids < 0).any():
        raise TypeError("Class ids must be positive integers")
    shape = list(ids.shape)
    shape[channel_dim] = num_classes
    logits = torch.zeros(shape, device=ids.device, dtype=torch.float32)
    logits.scatter_(channel_dim, ids.long(), 1)
    return logits


def convert_logits_to_ids(logits: torch.Tensor, *, channel_dim: int = 1) -> torch.Tensor:
    """
    :param logits: logits (..., num_classes, ...)
    :param channel_dim: the index of the channel dimension
    :return: class ids (..., 1, ...)
    """
    return logits.round().int() if logits.shape[channel_dim] < 2 else logits.argmax(channel_dim, keepdim=True)


def auto_convert(image: torch.Tensor) -> torch.Tensor:
    return (image * 255 if 0 <= image.min() <= image.max() <= 1 else Normalize(domain=(0, 255))(image)).int()



from abc import ABCMeta
from typing import override, Sequence

import numpy as np
import torch
from torch import nn, optim

from mipcandy.common import PolyLRScheduler, DiceBCELossWithLogits, DiceCELossWithLogits, Loss
from mipcandy.data import visualize2d, visualize3d, overlay, auto_convert, convert_logits_to_ids
from mipcandy.training import Trainer, TrainerToolbox
from mipcandy.types import Params


class DeepSupervisionWrapper(Loss):
    def __init__(self, loss: nn.Module, *, weight_factors: Sequence[float] | None = None) -> None:
        super().__init__()
        if weight_factors and all(x == 0 for x in weight_factors):
            raise ValueError("At least one weight factor should be nonzero")
        self.weight_factors: tuple[float, ...] | None = tuple(weight_factors) if weight_factors else None
        self.loss: nn.Module = loss

    def forward(self, outputs: Sequence[torch.Tensor], targets: Sequence[torch.Tensor]) -> tuple[
        torch.Tensor, dict[str, float]]:
        if not self.weight_factors:
            weights = (1.0,) * len(outputs)
        else:
            weights = self.weight_factors
        total_loss = torch.tensor(0, device=outputs[0].device, dtype=outputs[0].dtype)
        combined_metrics = {}
        for i, (output, target) in enumerate(zip(outputs, targets)):
            if weights[i] == 0:
                continue
            loss, metrics = self.loss(output, target)
            total_loss += weights[i] * loss
            for key, value in metrics.items():
                metric_key = f"{key}_ds{i}" if len(outputs) > 1 else key
                combined_metrics[metric_key] = value
        if combined_metrics:
            main_loss_key = next(iter(combined_metrics.keys())).replace("_ds0", "")
            if f"{main_loss_key}_ds0" in combined_metrics:
                combined_metrics[main_loss_key] = combined_metrics[f"{main_loss_key}_ds0"]
        return total_loss, combined_metrics


class SegmentationTrainer(Trainer, metaclass=ABCMeta):
    num_classes: int = 1
    include_background: bool = True
    deep_supervision: bool = False
    deep_supervision_scales: Sequence[float] | None = None
    deep_supervision_weights: Sequence[float] | None = None

    def _save_preview(self, x: torch.Tensor, title: str, quality: float, *, is_label: bool = False) -> None:
        path = f"{self.experiment_folder()}/{title} (preview).png"
        if x.ndim == 3 and x.shape[0] in (1, 3, 4):
            visualize2d(auto_convert(x), title=title, is_label=is_label, blocking=True, screenshot_as=path)
        elif x.ndim == 4 and x.shape[0] == 1:
            visualize3d(x, title=title, max_volume=int(quality * 1e6), is_label=is_label, blocking=True,
                        screenshot_as=path)

    def apply_non_linearity(self, x: torch.Tensor, channel_dim: int) -> torch.Tensor:
        return x.sigmoid() if self.num_classes < 2 else x.softmax(channel_dim)

    @override
    def save_preview(self, image: torch.Tensor, label: torch.Tensor, output: torch.Tensor, *,
                     quality: float = .75) -> None:
        output = convert_logits_to_ids(self.apply_non_linearity(output, 0), channel_dim=0)
        self._save_preview(image, "input", quality)
        self._save_preview(label.int(), "label", quality, is_label=True)
        self._save_preview(output, "prediction", quality, is_label=True)
        if image.ndim == label.ndim == output.ndim == 3 and label.shape[0] == output.shape[0] == 1:
            visualize2d(overlay(image, label), title="expected", blocking=True,
                        screenshot_as=f"{self.experiment_folder()}/expected (preview).png")
            visualize2d(overlay(image, output), title="actual", blocking=True,
                        screenshot_as=f"{self.experiment_folder()}/actual (preview).png")

    @override
    def build_ema(self, model: nn.Module) -> nn.Module:
        return optim.swa_utils.AveragedModel(model)

    @override
    def build_criterion(self) -> nn.Module:
        if self.num_classes < 2:
            if not self.include_background:
                raise ValueError("Binary segmentation models must include background class")
            loss = DiceBCELossWithLogits()
        else:
            loss = DiceCELossWithLogits(self.num_classes, include_background=self.include_background)
        if self.deep_supervision:
            if not self.deep_supervision_weights and self.deep_supervision_scales:
                weights = np.array([1 / (2 ** i) for i in range(len(self.deep_supervision_scales))])
                weights = weights / weights.sum()
                self.deep_supervision_weights = tuple(weights.tolist())
            loss = DeepSupervisionWrapper(loss, weight_factors=self.deep_supervision_weights)
            self.log(f"Deep supervision enabled with weights: {self.deep_supervision_weights}")
        return loss

    @override
    def build_optimizer(self, params: Params) -> optim.Optimizer:
        return optim.SGD(params, 1e-2, weight_decay=3e-5, momentum=.99, nesterov=True)

    @override
    def build_scheduler(self, optimizer: optim.Optimizer, num_epochs: int) -> optim.lr_scheduler.LRScheduler:
        return PolyLRScheduler(optimizer, 1e-2, num_epochs * len(self._dataloader))

    @override
    def backward(self, images: torch.Tensor, labels: torch.Tensor, toolbox: TrainerToolbox) -> tuple[float, dict[
        str, float]]:
        outputs = toolbox.model(images)
        if self.deep_supervision:
            if outputs.ndim == labels.ndim + 1:
                outputs = list(torch.unbind(outputs, dim=1))
            labels = self.prepare_deep_supervision_targets(labels, [m.shape[2:] for m in outputs])
        loss, metrics = toolbox.criterion(outputs, labels)
        loss.backward()
        nn.utils.clip_grad_norm_(toolbox.model.parameters(), 12)
        return loss.item(), metrics

    @staticmethod
    def prepare_deep_supervision_targets(labels: torch.Tensor, output_shapes: list[tuple[int, ...]]) -> list[
        torch.Tensor]:
        targets = []
        for shape in output_shapes:
            if labels.shape[2:] == shape:
                targets.append(labels)
            else:
                downsampled = nn.functional.interpolate(labels, shape,
                                                        mode="nearest-exact" if labels.ndim == 4 else "nearest")
                targets.append(downsampled)
        return targets

    def class_percentages(self, ids: torch.Tensor) -> dict[int, float]:
        bin_count = torch.bincount(ids.flatten(), minlength=self.num_classes)
        distribution = (bin_count / bin_count.sum()).cpu().tolist()
        return dict(enumerate(distribution))

    @staticmethod
    def format_class_percentages(percentages: dict[int, float], prefix: str) -> dict[str, float]:
        return {f"% {prefix} class {class_id}": percentage for class_id, percentage in percentages.items()}

    @override
    def validate_case(self, idx: int, image: torch.Tensor, label: torch.Tensor, toolbox: TrainerToolbox) -> tuple[
        float, dict[str, float], torch.Tensor]:
        image, label = image.unsqueeze(0), label.unsqueeze(0)
        output = (toolbox.ema if toolbox.ema else toolbox.model)(image)
        # the highest resolution is at index 0
        if isinstance(toolbox.criterion, Loss):
            toolbox.criterion.validation_mode = True
        if self.deep_supervision:
            if not isinstance(toolbox.criterion, DeepSupervisionWrapper):
                raise TypeError("Deep supervision is enabled but criterion is not a `DeepSupervisionWrapper`")
            if isinstance(output, (list, tuple)):
                output = output[0]
            elif output.ndim > label.ndim:
                output = output[:, 0]
            loss, metrics = toolbox.criterion([output], [label])
        else:
            loss, metrics = toolbox.criterion(output, label)
        if isinstance(toolbox.criterion, Loss):
            toolbox.criterion.validation_mode = False
        label_percentages = self.class_percentages(label)
        metrics.update(self.format_class_percentages(label_percentages, "label"))
        output_logits = self.apply_non_linearity(output, 1)
        output_percentages = self.class_percentages(convert_logits_to_ids(output_logits))
        metrics.update(self.format_class_percentages(output_percentages, "output"))
        return -loss.item(), metrics, output.squeeze(0)



from mipcandy.common.optim.loss import Loss, SegmentationLoss, FocalBCEWithLogits, DiceCELossWithLogits, \
    DiceBCELossWithLogits
from mipcandy.common.optim.lr_scheduler import AbsoluteLinearLR, PolyLRScheduler



from mipcandy.__entry__ import __entry__
from mipcandy.common import *
from mipcandy.config import load_settings, save_settings, load_secrets, save_secrets
from mipcandy.data import *
from mipcandy.evaluation import EvalCase, EvalResult, Evaluator
from mipcandy.frontend import *
from mipcandy.inference import parse_predictant, Predictor
from mipcandy.layer import batch_int_multiply, batch_int_divide, LayerT, HasDevice, auto_device, WithPaddingModule, \
    WithNetwork
from mipcandy.metrics import do_reduction, binary_dice, dice_similarity_coefficient, soft_dice
from mipcandy.presets import *
from mipcandy.profiler import ProfilerFrame, Profiler
from mipcandy.run import config
from mipcandy.sanity_check import num_trainable_params, model_complexity_info, SanityCheckResult, sanity_check
from mipcandy.training import TrainerToolbox, Trainer
from mipcandy.types import Setting, Settings, Params, Transform, SupportedPredictant, Colormap, Device, Shape2d, \
    Shape3d, Shape, AmbiguousShape, Paddings2d, Paddings3d, Paddings



from dataclasses import dataclass
from inspect import stack
from os import PathLike
from time import time
from typing import Sequence, override

import torch
from psutil import cpu_percent, virtual_memory

from mipcandy.data import dump_allocated_tensors
from mipcandy.types import Device, AmbiguousShape


@dataclass
class ProfilerFrame(object):
    stack: str
    cpu: float
    mem: float
    gpu: list[float] | None = None
    gpu_mem: list[float] | None = None

    @override
    def __str__(self) -> str:
        r = f"[{self.stack}] CPU: {self.cpu:.2f}% @ Memory: {self.mem:.2f}%\n"
        if self.gpu and self.gpu_mem:
            for i, gpu in enumerate(self.gpu):
                r += f"\t\tGPU {i}: {gpu:.2f}% @ Memory: {self.gpu_mem[i]:.2f}%\n"
        return r

    def export(self, duration: float) -> str:
        return f"{duration:.2f}s\t{self}"


class _LineBreak(object):
    def __init__(self, message: str) -> None:
        self.message: str = message

    @override
    def __str__(self) -> str:
        return f"<{self.message}>\n"

    def export(self, duration: float) -> str:
        return f"{duration:.2f}s\t{self}"


class Profiler(object):
    def __init__(self, title: str, save_as: str | PathLike[str], *, gpus: Sequence[Device] = ()) -> None:
        self.title: str = title
        self.save_as: str = save_as
        self.total_mem: float = self.get_total_mem()
        self.has_gpu: bool = len(gpus) > 0
        self._gpus: Sequence[Device] = gpus
        self.total_gpu_mem: list[float] = [self.get_total_gpu_mem(device) for device in gpus]
        with open(save_as, "w") as f:
            f.write(f"# {title}\nTotal memory: {self.total_mem}, Total GPU memory: {self.total_gpu_mem}\n\n")
        self._t0: float = time()
        self._allocated_tensors: tuple[float, list[tuple[
            float, AmbiguousShape, torch.dtype, torch.device, bool, str]]] = (0, [])

    @staticmethod
    def get_cpu_usage() -> float:
        return cpu_percent()

    def get_mem_usage(self) -> float:
        return 100 * virtual_memory().used / self.total_mem

    @staticmethod
    def get_total_mem() -> float:
        return virtual_memory().total

    @staticmethod
    def get_gpu_usage(device: Device) -> float:
        return torch.cuda.utilization(device)

    def get_gpu_mem_usage(self, device: Device) -> float:
        return 100 * torch.cuda.device_memory_used(device) / self.total_gpu_mem[self._gpus.index(device)]

    @staticmethod
    def get_total_gpu_mem(device: Device) -> float:
        return torch.cuda.get_device_properties(device).total_memory

    def _save(self, obj: ProfilerFrame | _LineBreak | str) -> None:
        with open(self.save_as, "a") as f:
            t = time()
            f.write(obj + "\n" if isinstance(obj, str) else f"{obj.export(t - self._t0)}\n")
            self._t0 = t

    def record_allocated_tensors(self, *, limit: int = 10) -> str:
        allocated_tensors = dump_allocated_tensors()
        counted_tensors = []
        added_tensors = []
        removed_tensors = []
        tensors, prev_tensors = allocated_tensors[1], self._allocated_tensors[1]
        for tensor in tensors + prev_tensors:
            if tensor in counted_tensors:
                continue
            sz, shape, dtype, device, requires_grad, grad_fn = tensor
            t = f"{sz:8.1f} MB | {shape} | {dtype} | {device} | grad={requires_grad} | {grad_fn}"
            if tensor in prev_tensors:
                if tensor in tensors:
                    num_diff = tensors.count(tensor) - prev_tensors.count(tensor)
                    if num_diff > 0:
                        added_tensors.append(f"{num_diff} x {t}")
                    if num_diff < 0:
                        removed_tensors.append(f"{num_diff} x {t}")
                else:
                    removed_tensors.append(f"{-prev_tensors.count(tensor)} x {t}")
            else:
                added_tensors.append(f"{tensors.count(tensor)} x {t}")
            counted_tensors.append(tensor)
        if len(added_tensors) > limit:
            added_tensors = added_tensors[:limit]
        if len(removed_tensors) > limit:
            removed_tensors = removed_tensors[:limit]
        r = (f"Total size diff: {allocated_tensors[0] - self._allocated_tensors[0]} MB\n"
             f"Added tensors:\n{"\t\n".join(added_tensors)}\nRemoved tensors:\n{"\t\n".join(removed_tensors)}\n")
        self._save(r)
        self._allocated_tensors = allocated_tensors
        return r

    def record(self, *, stack_trace_offset: int = 1) -> ProfilerFrame:
        frame = ProfilerFrame(" -> ".join([f"{f.function}:{f.lineno}" for f in reversed(stack()[stack_trace_offset:])]),
                              self.get_cpu_usage(), self.get_mem_usage())
        if self.has_gpu:
            frame.gpu = [torch.cuda.utilization(device) for device in self._gpus]
            frame.gpu_mem = [self.get_gpu_mem_usage(device) for device in self._gpus]
        self._save(frame)
        return frame

    def line_break(self, message: str) -> _LineBreak:
        r = _LineBreak(message)
        self._save(r)
        return r



from typing import Literal

import torch
from torch import nn

from mipcandy.data import convert_ids_to_logits, convert_logits_to_ids
from mipcandy.metrics import do_reduction, binary_dice, dice_similarity_coefficient, soft_dice


class FocalBCEWithLogits(nn.Module):
    def __init__(self, alpha: float, gamma: float, *, reduction: Literal["mean", "sum", "none"] = "mean") -> None:
        super().__init__()
        self.alpha: float = alpha
        self.gamma: float = gamma
        self.reduction: Literal["mean", "sum", "none"] = reduction

    def forward(self, logits: torch.Tensor, targets: torch.Tensor) -> torch.Tensor:
        bce = nn.functional.binary_cross_entropy_with_logits(logits, targets, reduction="none")
        p = torch.sigmoid(logits)
        p_t = torch.where(targets.bool(), p, 1 - p)
        alpha_t = torch.where(targets.bool(), torch.as_tensor(self.alpha, device=logits.device), torch.as_tensor(
            1 - self.alpha, device=logits.device))
        loss = alpha_t * (1 - p_t).pow(self.gamma) * bce
        return do_reduction(loss, self.reduction)


class Loss(nn.Module):
    def __init__(self) -> None:
        super().__init__()
        self._validation_mode: bool = False

    @property
    def validation_mode(self) -> bool:
        return self._validation_mode

    @validation_mode.setter
    def validation_mode(self, value: bool) -> None:
        self._validation_mode = value
        for child in self.children():
            if isinstance(child, Loss):
                child.validation_mode = value


class SegmentationLoss(Loss):
    def __init__(self, num_classes: int, include_background: bool) -> None:
        super().__init__()
        self.num_classes: int = num_classes
        self.include_background: bool = include_background

    def logitfy_no_grad(self, ids: torch.Tensor) -> torch.Tensor:
        if self.num_classes != 1 and ids.shape[1] == 1:
            with torch.no_grad():
                return convert_ids_to_logits(ids.int(), self.num_classes)
        return ids.float()


class DiceCELossWithLogits(SegmentationLoss):
    def __init__(self, num_classes: int, *, lambda_ce: float = 1, lambda_soft_dice: float = 1,
                 smooth: float = 1e-5, include_background: bool = True) -> None:
        super().__init__(num_classes, include_background)
        self.lambda_ce: float = lambda_ce
        self.lambda_soft_dice: float = lambda_soft_dice
        self.smooth: float = smooth

    def _forward(self, outputs: torch.Tensor, labels: torch.Tensor) -> tuple[torch.Tensor, dict[str, float]]:
        ce = nn.functional.cross_entropy(outputs, labels[:, 0].long())
        outputs = outputs.softmax(1)
        labels = self.logitfy_no_grad(labels)
        if not self.include_background:
            outputs = outputs[:, 1:]
            labels = labels[:, 1:]
        dice = soft_dice(outputs, labels, smooth=self.smooth)
        metrics = {"soft dice": dice.item(), "ce loss": ce.item()}
        c = self.lambda_ce * ce + self.lambda_soft_dice * (1 - dice)
        return c, metrics

    def forward(self, outputs: torch.Tensor, labels: torch.Tensor) -> tuple[torch.Tensor, dict[str, float]]:
        if not self.validation_mode:
            return self._forward(outputs, labels)
        with torch.no_grad():
            c, metrics = self._forward(outputs, labels)
            outputs = convert_logits_to_ids(outputs)
            for i in range(0 if self.include_background else 1, self.num_classes):
                class_dice = binary_dice(outputs == i, labels == i).item()
                metrics[f"dice {i}"] = class_dice
            metrics["dice"] = dice_similarity_coefficient(
                self.logitfy_no_grad(outputs), self.logitfy_no_grad(labels)
            ).item()
            return c, metrics


class DiceBCELossWithLogits(SegmentationLoss):
    def __init__(self, *, lambda_bce: float = 1, lambda_soft_dice: float = 1,
                 smooth: float = 1e-5, min_percentage_per_class: float | None = None) -> None:
        super().__init__(1, True)
        self.lambda_bce: float = lambda_bce
        self.lambda_soft_dice: float = lambda_soft_dice
        self.smooth: float = smooth
        self.min_percentage_per_class: float | None = min_percentage_per_class

    def _forward(self, outputs: torch.Tensor, labels: torch.Tensor) -> tuple[torch.Tensor, dict[str, float]]:
        outputs = outputs.sigmoid()
        labels = labels.float()
        bce = nn.functional.binary_cross_entropy(outputs, labels)
        dice = soft_dice(outputs, labels, smooth=self.smooth)
        metrics = {"soft dice": dice.item(), "bce loss": bce.item()}
        c = self.lambda_bce * bce + self.lambda_soft_dice * (1 - dice)
        return c, metrics

    def forward(self, outputs: torch.Tensor, labels: torch.Tensor) -> tuple[torch.Tensor, dict[str, float]]:
        if not self.validation_mode:
            return self._forward(outputs, labels)
        with torch.no_grad():
            c, metrics = self._forward(outputs, labels)
            outputs = convert_logits_to_ids(outputs).bool()
            metrics["dice"] = binary_dice(outputs, labels.bool()).item()
            return c, metrics



from abc import ABCMeta
from math import log, ceil
from os import PathLike, listdir
from os.path import isdir, basename, exists
from typing import Sequence

import torch
from torch import nn

from mipcandy.data import save_image, Loader, UnsupervisedDataset, PathBasedUnsupervisedDataset
from mipcandy.layer import WithPaddingModule, WithNetwork
from mipcandy.types import SupportedPredictant, Device, AmbiguousShape


def parse_predictant(x: SupportedPredictant, loader: type[Loader], *, as_label: bool = False) -> tuple[list[
    torch.Tensor], list[str] | None]:
    if isinstance(x, str):
        if isdir(x):
            cases = listdir(x)
            return [loader.do_load(f"{x}/{case}", is_label=as_label) for case in cases], cases
        return [loader.do_load(x, is_label=as_label)], [basename(x)]
    if isinstance(x, torch.Tensor):
        return [x], None
    r, filenames = [], None
    for case in x:
        if isinstance(case, str):
            if not filenames:
                filenames = []
            r.append(loader.do_load(case, is_label=as_label))
            filenames.append(case[case.rfind("/") + 1:])
        elif filenames:
            raise TypeError("`x` should be single-typed")
        elif isinstance(case, torch.Tensor):
            r.append(case)
        else:
            raise TypeError(f"Unexpected type of element {type(case)}")
    return r, filenames


class Predictor(WithPaddingModule, WithNetwork, metaclass=ABCMeta):
    def __init__(self, experiment_folder: str | PathLike[str], example_shape: AmbiguousShape, *,
                 checkpoint: str = "checkpoint_best.pth", device: Device = "cpu") -> None:
        WithPaddingModule.__init__(self, device)
        WithNetwork.__init__(self, device)
        self._experiment_folder: str = experiment_folder
        self._example_shape: AmbiguousShape = example_shape
        self._checkpoint: str = checkpoint
        self._model: nn.Module | None = None

    def lazy_load_model(self) -> None:
        if self._model:
            return
        self._model = self.load_model(self._example_shape, False,
                                      path=f"{self._experiment_folder}/{self._checkpoint}")
        self._model.eval()

    def predict_image(self, image: torch.Tensor, *, batch: bool = False) -> torch.Tensor:
        self.lazy_load_model()
        image = image.to(self._device)
        if not batch:
            image = image.unsqueeze(0)
        padding_module = self.get_padding_module()
        if padding_module:
            image = padding_module(image)
        output = self._model(image)
        restoring_module = self.get_restoring_module()
        if restoring_module:
            output = restoring_module(output)
        return output if batch else output.squeeze(0)

    def _predict(self, x: SupportedPredictant | UnsupervisedDataset) -> tuple[list[torch.Tensor], list[str] | None]:
        if isinstance(x, PathBasedUnsupervisedDataset):
            return [self.predict_image(case) for case in x], x.paths()
        if isinstance(x, UnsupervisedDataset):
            return [self.predict_image(case) for case in x], None
        images, filenames = parse_predictant(x, Loader)
        return [self.predict_image(image) for image in images], filenames

    def predict(self, x: SupportedPredictant | UnsupervisedDataset) -> list[torch.Tensor]:
        return self._predict(x)[0]

    @staticmethod
    def save_prediction(output: torch.Tensor, path: str | PathLike[str]) -> None:
        save_image(output, path)

    def save_predictions(self, outputs: Sequence[torch.Tensor], folder: str | PathLike[str], *,
                         filenames: Sequence[str | PathLike[str]] | None = None) -> None:
        if not exists(folder):
            raise FileNotFoundError(f"Folder {folder} does not exist")
        if not filenames:
            num_digits = ceil(log(len(outputs)))
            filenames = [f"prediction_{str(i).zfill(num_digits)}.{
            "png" if output.ndim == 3 and output.shape[0] in (1, 3) else "mha"}" for i, output in enumerate(outputs)]
        for i, prediction in enumerate(outputs):
            self.save_prediction(prediction, f"{folder}/{filenames[i]}")

    def predict_to_files(self, x: SupportedPredictant | UnsupervisedDataset,
                         folder: str | PathLike[str]) -> list[str] | None:
        outputs, filenames = self._predict(x)
        self.save_predictions(outputs, folder, filenames=filenames)
        return filenames

    def __call__(self, x: SupportedPredictant | UnsupervisedDataset) -> list[torch.Tensor]:
        return self.predict(x)



from mipcandy.common.module.conv import ConvBlock2d, ConvBlock3d, WSConv2d, WSConv3d
from mipcandy.common.module.preprocess import Pad2d, Pad3d, Restore2d, Restore3d, PadTo, Normalize, CTNormalize, \
    ColorizeLabel



import torch

from mipcandy.types import Device, Reduction


def _args_check(outputs: torch.Tensor, labels: torch.Tensor, *, dtype: torch.dtype | None = None,
                device: Device | None = None) -> tuple[torch.dtype, Device]:
    if outputs.shape != labels.shape:
        raise ValueError(f"Outputs ({outputs.shape}) and labels ({labels.shape}) must have the same shape")
    if (outputs_dtype := outputs.dtype) != labels.dtype or dtype and outputs_dtype != dtype:
        raise TypeError(f"Outputs({outputs_dtype}) and labels ({labels.dtype}) must both be {dtype}")
    if (outputs_device := outputs.device) != labels.device:
        raise RuntimeError(f"Outputs ({outputs.device}) and labels ({labels.device}) must be on the same device")
    if device and outputs_device != device:
        raise RuntimeError(f"Tensors are expected to be on {device}, but instead they are on {outputs.device}")
    return outputs_dtype, outputs_device


def do_reduction(x: torch.Tensor, method: Reduction) -> torch.Tensor:
    match method:
        case "mean":
            return x.mean()
        case "median":
            return x.median()
        case "sum":
            return x.sum()
        case "none":
            return x


def binary_dice(outputs: torch.Tensor, labels: torch.Tensor, *, if_empty: float = 1,
                reduction: Reduction = "mean") -> torch.Tensor:
    """
    :param outputs: boolean class ids (B, 1, ...)
    :param labels: boolean class ids (B, 1, ...)
    :param if_empty: the value to return if both outputs and labels are empty
    :param reduction: the reduction method to apply to the dice score
    """
    _args_check(outputs, labels, dtype=torch.bool)
    axes = tuple(range(2, outputs.ndim))
    volume_sum = outputs.sum(axes) + labels.sum(axes)
    if volume_sum == 0:
        return torch.tensor(if_empty, dtype=torch.float)
    return do_reduction(2 * (outputs & labels).sum(axes) / volume_sum, reduction)


def dice_similarity_coefficient(outputs: torch.Tensor, labels: torch.Tensor, *, if_empty: float = 1,
                                reduction: Reduction = "mean") -> torch.Tensor:
    """
    :param outputs: one-hot (B, N, ...)
    :param labels: one-hot (B, N, ...)
    :param if_empty: the value to return if both outputs and labels are empty
    :param reduction: the reduction method to apply to the dice score
    """
    _args_check(outputs, labels, dtype=torch.float)
    axes = tuple(range(2, outputs.ndim))
    tp = (outputs * labels).sum(axes)
    fp = (outputs * (1 - labels)).sum(axes)
    fn = ((1 - outputs) * labels).sum(axes)
    volume_sum = 2 * tp + fp + fn
    if (volume_sum == 0).any():
        return torch.tensor(if_empty, dtype=torch.float)
    return do_reduction(2 * tp / volume_sum, reduction)


def soft_dice(outputs: torch.Tensor, labels: torch.Tensor, *, smooth: float = 1, batch_dice: bool = True,
              reduction: Reduction = "mean") -> torch.Tensor:
    """
    :param outputs: logits (B, C, ...)
    :param labels: logits (B, C, ...)
    :param smooth: the smoothness term to avoid division by zero
    :param batch_dice: whether to compute dice score for each batch separately
    :param reduction: the reduction method to apply to the dice score
    """
    _args_check(outputs, labels, dtype=torch.float)
    axes = tuple(range(2, outputs.ndim))
    if batch_dice:
        axes = (0,) + axes
    label_sum = labels.sum(axes)
    intersection = (outputs * labels).sum(axes)
    output_sum = outputs.sum(axes)
    return do_reduction((2 * intersection + smooth) / (label_sum + output_sum + smooth), reduction)



from importlib.util import find_spec
from typing import override

from mipcandy.frontend.prototype import Frontend
from mipcandy.types import Settings

if find_spec("wandb"):
    from wandb import init, Run


    class WandBFrontend(Frontend):
        def __init__(self, secrets: Settings) -> None:
            super().__init__(secrets)
            self._entity: str = self.require_nonempty_secret("wandb_entity", required_type=str)
            self._project: str = self.require_nonempty_secret("wandb_project", required_type=str)
            self._run: Run | None = None

        @override
        def on_experiment_created(self, experiment_id: str, trainer: str, model: str, note: str, num_macs: float,
                                  num_params: float, num_epochs: int, early_stop_tolerance: int) -> None:
            self._run = init(entity=self._entity, project=self._project, config={
                "experiment_id": experiment_id, "trainer": trainer, "model": model, "note": note, "num_macs": num_macs,
                "num_params": num_params, "num_epochs": num_epochs
            })

        @override
        def on_experiment_updated(self, experiment_id: str, epoch: int, metrics: dict[str, list[float]],
                                  early_stop_tolerance: int) -> None:
            if self._run:
                self._run.log(metrics)

        @override
        def on_experiment_completed(self, experiment_id: str) -> None:
            if not self._run:
                raise RuntimeError("Experiment has not been created")
            self._run.finish()
else:
    WandBFrontend: type[Frontend] = Frontend
